/*
** Module   : WidgetBrowserAspect.java
** Abstract : The integration of Widget Browser into the widget drawing infrastructure.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141031 Created initial version
*/
package com.goldencode.p2j.ui.client.widgetbrowser;

import java.awt.Color;
import java.awt.event.*;
import java.util.*;

import org.aspectj.lang.annotation.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.ScreenDriver;
import com.goldencode.p2j.ui.client.gui.driver.GuiDriver;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This aspect is responsible for lifetime management of the Widget Browser component
 * and the integration with the widget drawing facility.
 * <p>
 * To activate the Widget Browser feature the Java property "widgetbrowser" must be defined,
 * for example using the -D option for the Java application launcher. 
 * The class displays the Widget Browser main window for each drawn {@link Window}.  
 */
@Aspect
public class WidgetBrowserAspect
{
   /**
    * Map of active Widget Browsers.
    */
   private Map<Window<?>, WidgetBrowser> activeBrowsers = new HashMap<>();
   
   /**
    * This method is called on every {@link Widget#draw()} method and serves as the main
    * integration entry-point into the widget drawing facility.
    * <p>
    * For production-quality code, this class can be omitted from AdpectJ compilation to
    * remove the Widget Browser overhead.
    * 
    * @param   widget
    *          The target on which the {@link Widget#draw()} method is called.
    */
   @AfterReturning("execution(void com.goldencode.p2j.ui.client.widget.Widget.draw()) && " +
                   "target(widget)")
   public void onWidgetDrawn(@SuppressWarnings("rawtypes") Widget widget)
   {
      // is widget browser active?
      if (System.getProperty("widgetbrowser") == null)
      {
         return;
      }
      
      // take care of starting new widget browser
      if (widget instanceof Window<?>)
      {
         handleStartWidgetBrowser((Window<?>) widget);
      }
      
      // the widget may need the highlight
      handleHighlightWidget(widget);
   }
   
   /**
    * The method creates and shows new Widget Browser main window.
    * If one has been created for the supplied window, this method is a no-op.
    * 
    * @param   window
    *          The {@link Window} instance for which new Widget Browser is being started.
    */
   private synchronized void handleStartWidgetBrowser(Window<?> window)
   {
      if (activeBrowsers.containsKey(window))
      {
         // browser for this window already exists, nothing to do
         return;
      }
      
      WidgetBrowser browserWindow = new WidgetBrowser((Container<?>) window, new ActionListener() 
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            switch (e.getID())
            {
            case Event.WIDGET_SELECTION_CHANGED:
               highlightWidget(((Event) e).widget);
               break;
            case Event.BROWSER_CLOSED:
               ScreenDriver<?> driver = ((Event) e).widget.screen().getInstanceDriver();
               if (driver instanceof GuiDriver)
               {
                  ((GuiDriver) driver).startBatch();
                  try
                  {
                     ((GuiDriver) driver).clearHighlight();
                  }
                  finally
                  {
                     ((GuiDriver) driver).endBatch();
                  }
               }
               break;
            }
         }
      });

      browserWindow.show();
      activeBrowsers.put(window, browserWindow);
   }
   
   /**
    * Highlights the supplied widget when a related active Widget Browser exists, 
    * the widget is visible and it is selected to be highlighted in the Widget Browser.
    * 
    * @param   widget
    *          The widget to be highlighted.
    */
   private synchronized void handleHighlightWidget(Widget<?> widget)
   {
      Window<?> window = (Window<?>) widget.ancestor();
      if (window == null)
      {
         return;
      }
      
      WidgetBrowser browser = activeBrowsers.get(window);
      if (browser == null || !browser.isHighlighted(widget))
      {
         return;
      }
      
      highlightWidget(widget);
   }
   
   /**
    * Highlights the supplied widget when the widget is visible.
    * 
    * @param   widget
    *          The widget to be highlighted.
    */
   private void highlightWidget(Widget<?> widget)
   {
      if (widget == null)
      {
         return;
      }
      
      OutputManager<?> om = widget.screen();
      if (om.getInstanceDriver() instanceof GuiDriver)
      {
         GuiDriver driver = ((GuiDriver) om.getInstanceDriver());
         
         // draw in a batch so that the widget is redrawn
         driver.startBatch();
         try 
         {
            driver.clearHighlight();
            
            if (widget.isVisible())
            {
               NativePoint loc = widget.physicalLocation();
               NativeDimension dim = widget.physicalDimension();
               driver.drawHighlight(loc.x, loc.y, dim.width, dim.height, Color.CYAN);
            }
         }
         finally
         {
            driver.endBatch();
         }
      }
   }
}